package org.bidib.wizard.mvc.main.controller;

import gnu.io.PortInUseException;
import gnu.io.UnsupportedCommOperationException;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import javax.swing.ImageIcon;
import javax.swing.JList;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.bidib.jbidibc.Bidib;
import org.bidib.jbidibc.Feature;
import org.bidib.jbidibc.MessageListener;
import org.bidib.jbidibc.MessageReceiver;
import org.bidib.jbidibc.enumeration.BoosterState;
import org.bidib.jbidibc.enumeration.IdentifyState;
import org.bidib.jbidibc.exception.PortNotFoundException;
import org.bidib.jbidibc.exception.ProtocolException;
import org.bidib.jbidibc.node.NodeFactory;
import org.bidib.jbidibc.node.RootNode;
import org.bidib.jbidibc.node.listener.TransferListener;
import org.bidib.wizard.comm.AnalogPortStatus;
import org.bidib.wizard.comm.BoosterStatus;
import org.bidib.wizard.comm.Communication;
import org.bidib.wizard.comm.CommunicationFactory;
import org.bidib.wizard.comm.FeedbackPortStatus;
import org.bidib.wizard.comm.LightPortStatus;
import org.bidib.wizard.comm.MotorPortStatus;
import org.bidib.wizard.comm.SoundPortStatus;
import org.bidib.wizard.comm.SwitchPortStatus;
import org.bidib.wizard.comm.listener.CommunicationListener;
import org.bidib.wizard.dialog.AboutDialog;
import org.bidib.wizard.dialog.FileDialog;
import org.bidib.wizard.dialog.NodeDetailsDialog;
import org.bidib.wizard.locale.Resources;
import org.bidib.wizard.mvc.cvprogrammer.controller.CvProgrammerController;
import org.bidib.wizard.mvc.cvprogrammer.controller.listener.CvProgrammerControllerListener;
import org.bidib.wizard.mvc.features.controller.FeaturesController;
import org.bidib.wizard.mvc.features.controller.listener.FeaturesControllerListener;
import org.bidib.wizard.mvc.firmware.controller.FirmwareController;
import org.bidib.wizard.mvc.firmware.controller.listener.FirmwareControllerListener;
import org.bidib.wizard.mvc.loco.controller.LocoController;
import org.bidib.wizard.mvc.main.model.Accessory;
import org.bidib.wizard.mvc.main.model.AccessoryLabels;
import org.bidib.wizard.mvc.main.model.AddressData;
import org.bidib.wizard.mvc.main.model.AnalogPort;
import org.bidib.wizard.mvc.main.model.AnalogPortLabels;
import org.bidib.wizard.mvc.main.model.FeedbackPort;
import org.bidib.wizard.mvc.main.model.FeedbackPortLabels;
import org.bidib.wizard.mvc.main.model.Flag;
import org.bidib.wizard.mvc.main.model.InputPort;
import org.bidib.wizard.mvc.main.model.InputPortLabels;
import org.bidib.wizard.mvc.main.model.LightPort;
import org.bidib.wizard.mvc.main.model.LightPortLabels;
import org.bidib.wizard.mvc.main.model.Macro;
import org.bidib.wizard.mvc.main.model.MacroLabels;
import org.bidib.wizard.mvc.main.model.MainModel;
import org.bidib.wizard.mvc.main.model.MotorPort;
import org.bidib.wizard.mvc.main.model.MotorPortLabels;
import org.bidib.wizard.mvc.main.model.Node;
import org.bidib.wizard.mvc.main.model.NodeLabels;
import org.bidib.wizard.mvc.main.model.NodeState;
import org.bidib.wizard.mvc.main.model.Port;
import org.bidib.wizard.mvc.main.model.ServoPort;
import org.bidib.wizard.mvc.main.model.ServoPortLabels;
import org.bidib.wizard.mvc.main.model.SoundPort;
import org.bidib.wizard.mvc.main.model.SoundPortLabels;
import org.bidib.wizard.mvc.main.model.SwitchPort;
import org.bidib.wizard.mvc.main.model.SwitchPortLabels;
import org.bidib.wizard.mvc.main.model.function.AnalogPortAction;
import org.bidib.wizard.mvc.main.model.function.FlagFunction;
import org.bidib.wizard.mvc.main.model.function.Function;
import org.bidib.wizard.mvc.main.model.function.InputFunction;
import org.bidib.wizard.mvc.main.model.function.LightPortAction;
import org.bidib.wizard.mvc.main.model.function.MacroFunction;
import org.bidib.wizard.mvc.main.model.function.MotorPortAction;
import org.bidib.wizard.mvc.main.model.function.PortAction;
import org.bidib.wizard.mvc.main.model.function.ServoPortAction;
import org.bidib.wizard.mvc.main.model.function.SoundPortAction;
import org.bidib.wizard.mvc.main.model.function.SwitchPortAction;
import org.bidib.wizard.mvc.main.model.listener.LightPortListener;
import org.bidib.wizard.mvc.main.model.listener.OutputListener;
import org.bidib.wizard.mvc.main.model.listener.PortListener;
import org.bidib.wizard.mvc.main.model.listener.ServoPortListener;
import org.bidib.wizard.mvc.main.view.MainView;
import org.bidib.wizard.mvc.main.view.menu.listener.MainMenuListener;
import org.bidib.wizard.mvc.main.view.panel.listener.AccessoryListListener;
import org.bidib.wizard.mvc.main.view.panel.listener.AccessoryTableListener;
import org.bidib.wizard.mvc.main.view.panel.listener.MacroListListener;
import org.bidib.wizard.mvc.main.view.panel.listener.MacroTableListener;
import org.bidib.wizard.mvc.main.view.panel.listener.NodeListListener;
import org.bidib.wizard.mvc.main.view.panel.listener.StatusListener;
import org.bidib.wizard.mvc.preferences.controller.PreferencesController;
import org.bidib.wizard.mvc.preferences.model.Preferences;
import org.bidib.wizard.mvc.preferences.model.listener.PreferencesListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class MainController {
    private static final Logger LOGGER = LoggerFactory.getLogger(MainController.class);

    // description, suffix for accessory files
    private static String accessoryDescription;

    private static final String ACCESSORY_EXTENSION = "accessory";

    private static FileFilter accessoryFilter;

    // description, suffix for macro files
    private static String macroDescription;

    private static final String MACRO_EXTENSION = "macro";

    private static FileFilter macroFilter;

    // description, suffix for node files
    private static String nodeDescription;

    private static final String NODE_EXTENSION = "node";

    private static FileFilter nodeFilter;

    private final MainModel model = new MainModel();

    private MainView view;

    private final AccessoryLabels accessoryLabels = new AccessoryLabels();

    private final AnalogPortLabels analogPortLabels = new AnalogPortLabels();

    private final FeedbackPortLabels feedbackPortLabels = new FeedbackPortLabels();

    private final InputPortLabels inputPortLabels = new InputPortLabels();

    private final LightPortLabels lightPortLabels = new LightPortLabels();

    private final MacroLabels macroLabels = new MacroLabels();

    private final MotorPortLabels motorPortLabels = new MotorPortLabels();

    private final NodeLabels nodeLabels = new NodeLabels();

    private final ServoPortLabels servoPortLabels = new ServoPortLabels();

    private final SoundPortLabels soundPortLabels = new SoundPortLabels();

    private final SwitchPortLabels switchPortLabels = new SwitchPortLabels();

    private Function[] functions;

    public MainController() {
        accessoryDescription = Resources.getString(MainController.class, "accessoryDescription");
        accessoryFilter = new FileNameExtensionFilter(accessoryDescription, ACCESSORY_EXTENSION);
        macroDescription = Resources.getString(MainController.class, "macroDescription");
        macroFilter = new FileNameExtensionFilter(macroDescription, MACRO_EXTENSION);
        nodeDescription = Resources.getString(MainController.class, "nodeDescription");
        nodeFilter = new FileNameExtensionFilter(nodeDescription, NODE_EXTENSION);
    }

    private Node createNode(Communication communication, org.bidib.jbidibc.Node node) {
        Node result = null;
        LOGGER.debug("Create new 'wizard' node from jbidibc.Node: {}", node);

        if (communication != null && node != null) {
            result = new Node();
            communication.clearCachedData(node);

            // collect some configuration of the node
            result.setCommandStation(communication.isCommandStation(node));

            // if the node has feedback ports we know its a feedback device
            int numFeedbackPorts = communication.getFeedbackPorts(node).size();
            LOGGER.debug("Number of feedback ports: {}", numFeedbackPorts);
            if (numFeedbackPorts > 0) {
                LOGGER.debug("The new node has feedback ports.");
                result.setAddressMessagesEnabled(communication.isAddressMessagesEnabled(node));
                result.setFeedbackMessagesEnabled(communication.isFeedbackMessagesEnabled(node));
            }
            int numInputPorts = communication.getInputPorts(node).size();
            LOGGER.debug("Number of input ports: {}", numInputPorts);
            if (numInputPorts > 0) {
                LOGGER.debug("The new node has input ports.");
                result.setKeyMessagesEnabled(communication.isKeyMessagesEnabled(node));
            }
            int macroLength = communication.getMacroLength(node);
            LOGGER.debug("Supported macro lenght: {}", macroLength);
            if (macroLength > 0) {
                LOGGER.debug("The new node has macros.");
                result.setDccStartEnabled(communication.isDccStartEnabled(node));
                result.setExternalStartEnabled(communication.isExternalStartEnabled(node));
            }
            result.setLabel(nodeLabels.getLabel(node.getUniqueId()));
            result.setNode(node);
            result.setStorableMacroCount(communication.getStorableMacroCount(node));
            // check if the node supports FW update
            result.setUpdatable(communication.isUpdatable(node));
        }
        return result;
    }

    private void createNodes(boolean mock) {
        LOGGER.debug("Create nodes, mock: {}", mock);
        model.getStatusModel().setCd(false);
        model.getStatusModel().setRx(false);
        model.getStatusModel().setTx(false);

        if (!mock) {
            model.getStatusModel().setCd(true);
        }

        Collection<Node> nodes = new LinkedList<Node>();
        Communication communication = CommunicationFactory.getInstance();

        if (communication != null) {

            // iterate over all nodes registered in the interface
            for (org.bidib.jbidibc.Node commNode : communication.getNodes()) {
                Node node = createNode(communication, commNode);

                if (node != null) {
                    nodes.add(node);
                }
            }
            communication.enable(null);
        }
        else {
            LOGGER.info("Communication factory is not available.");
            model.getStatusModel().setCd(false);
        }
        model.setNodes(nodes);
    }

    private Node findNode(org.bidib.jbidibc.Node commNode) {
        Node result = null;

        if (commNode != null) {
            for (Node node : model.getNodes()) {
                if (node.getNode().getUniqueId() == commNode.getUniqueId()) {
                    result = node;
                    break;
                }
            }
        }
        return result;
    }

    private Node findNode(byte[] address) {
        Node result = null;

        if (address != null) {
            int addressValue = NodeFactory.convert(address);

            for (Node node : model.getNodes()) {
                if (addressValue == NodeFactory.convert(node.getNode().getAddr())) {
                    result = node;
                    break;
                }
            }
        }
        return result;
    }

    private Function[] getFunctions() {
        Function[] result = null;

        if (functions != null) {
            try {
                result = new Function[functions.length];
                for (int index = 0; index < functions.length; index++) {
                    if (functions[index] != null) {
                        result[index] = ((Function) functions[index].clone());
                    }
                }
            }
            catch (CloneNotSupportedException e) {
                throw new RuntimeException(e);
            }
        }
        return result;
    }

    private void loadLabels() {
        accessoryLabels.load();
        analogPortLabels.load();
        feedbackPortLabels.load();
        inputPortLabels.load();
        lightPortLabels.load();
        macroLabels.load();
        motorPortLabels.load();
        nodeLabels.load();
        servoPortLabels.load();
        soundPortLabels.load();
        switchPortLabels.load();
    }

    private void replaceReferences(Macro macro) {
        if (macro != null) {
            Collection<Function> macroFunctions = macro.getFunctions();

            if (macroFunctions != null) {
                for (Function function : macroFunctions) {
                    if (function instanceof AnalogPortAction) {
                        Port port = ((PortAction) function).getPort();

                        if (port != null) {
                            int id = port.getId();
                            List<? extends Port> ports = model.getAnalogPorts();

                            if (id >= 0 && id < ports.size()) {
                                ((PortAction) function).setPort(ports.get(id));
                            }
                        }
                    }
                    else if (function instanceof FlagFunction) {
                        Flag flag = ((FlagFunction) function).getFlag();

                        if (flag != null) {
                            int id = flag.getId();
                            List<Flag> flags = model.getFlags();

                            if (id >= 0 && id < flags.size()) {
                                ((FlagFunction) function).setFlag(flags.get(id));
                            }
                        }
                    }
                    else if (function instanceof InputFunction) {
                        Port port = ((InputFunction) function).getInput();

                        if (port != null) {
                            int id = port.getId();
                            List<InputPort> ports = model.getInputPorts();

                            if (id >= 0 && id < ports.size()) {
                                ((InputFunction) function).setInput(ports.get(id));
                            }
                        }
                    }
                    else if (function instanceof LightPortAction) {
                        Port port = ((PortAction) function).getPort();

                        if (port != null) {
                            int id = port.getId();
                            List<? extends Port> ports = model.getLightPorts();

                            if (id >= 0 && id < ports.size()) {
                                ((PortAction) function).setPort(ports.get(id));
                            }
                        }
                    }
                    else if (function instanceof MacroFunction) {
                        ((MacroFunction) function).setMacroId(((MacroFunction) function).getMacroId());
                    }
                    else if (function instanceof MotorPortAction) {
                        Port port = ((PortAction) function).getPort();

                        if (port != null) {
                            int id = port.getId();
                            List<? extends Port> ports = model.getMotorPorts();

                            if (id >= 0 && id < ports.size()) {
                                ((PortAction) function).setPort(ports.get(id));
                            }
                        }
                    }
                    else if (function instanceof ServoPortAction) {
                        Port port = ((PortAction) function).getPort();

                        if (port != null) {
                            int id = port.getId();
                            List<? extends Port> ports = model.getServoPorts();

                            if (id >= 0 && id < ports.size()) {
                                ((PortAction) function).setPort(ports.get(id));
                            }
                        }
                    }
                    else if (function instanceof SoundPortAction) {
                        Port port = ((PortAction) function).getPort();

                        if (port != null) {
                            int id = port.getId();
                            List<? extends Port> ports = model.getSoundPorts();

                            if (id >= 0 && id < ports.size()) {
                                ((PortAction) function).setPort(ports.get(id));
                            }
                        }
                    }
                    else if (function instanceof SwitchPortAction) {
                        Port port = ((PortAction) function).getPort();

                        if (port != null) {
                            int id = port.getId();
                            List<? extends Port> ports = model.getSwitchPorts();

                            if (id >= 0 && id < ports.size()) {
                                ((PortAction) function).setPort(ports.get(id));
                            }
                        }
                    }
                }
            }
        }
    }

    private void setDefaultCursor() {
        view.setBusy(false);
    }

    private void setFunctions(Function[] functions) {
        if (functions != null) {
            try {
                this.functions = new Function[functions.length];
                for (int index = 0; index < functions.length; index++) {
                    if (functions[index] != null) {
                        this.functions[index] = ((Function) functions[index].clone());
                    }
                }
            }
            catch (CloneNotSupportedException e) {
                throw new RuntimeException(e);
            }
        }
        else {
            this.functions = null;
        }
    }

    private void setWaitCursor() {
        view.setBusy(true);
    }

    /**
     * Start the main controller
     * 
     * @throws PortNotFoundException
     * @throws PortInUseException
     * @throws UnsupportedCommOperationException
     * @throws IOException
     * @throws ProtocolException
     */
    public void start() throws PortNotFoundException, PortInUseException, UnsupportedCommOperationException,
        IOException, ProtocolException {

        LOGGER.info("Start the main controller");

        loadLabels();
        RootNode rootNode = Bidib.getInstance().getRootNode();
        rootNode.addTransferListener(new TransferListener() {
            @Override
            public void receiveStarted() {
                model.getStatusModel().setRx(true);
            }

            @Override
            public void receiveStopped() {
                model.getStatusModel().setRx(false);
            }

            @Override
            public void sendStarted() {
                model.getStatusModel().setTx(true);
            }

            @Override
            public void sendStopped() {
                model.getStatusModel().setTx(false);
            }
        });

        MessageReceiver messageReceiver = Bidib.getInstance().getMessageReceiver();
        messageReceiver.addMessageListener(new MessageListener() {
            @Override
            public void address(
                byte[] address, int detectorNumber, Collection<org.bidib.jbidibc.AddressData> addressDatas) {
                LOGGER.debug("address message, address: {}, detectorNumber: {}, addressDatas: {}", address,
                    detectorNumber, addressDatas);
                Node node = findNode(address);

                if (model.getNode() != null && model.getNode().equals(node)) {
                    Collection<AddressData> addresses = new LinkedList<AddressData>();

                    for (org.bidib.jbidibc.AddressData addressData : addressDatas) {
                        addresses.add(new AddressData(addressData.getAddress(), addressData.getType()));
                    }
                    model.setFeedbackPortAddresses(detectorNumber, addresses);
                }
            }

            @Override
            public void boosterCurrent(byte[] address, int current) {
                Node node = findNode(address);

                if (model.getNode() != null && model.getNode().equals(node)) {
                    LOGGER.trace("boosterCurrent, address: {}, current: {}mA", address, current);
                    model.setBoosterCurrent(current);
                }
            }

            @Override
            public void boosterState(byte[] address, BoosterState boosterState) {
                Node node = findNode(address);

                if (model.getNode() != null && model.getNode().equals(node)) {
                    LOGGER.trace("Booster state has changed, address: {}, state: {}", address, boosterState);
                    model.setBoosterStatus(BoosterStatus.valueOf(boosterState));
                }
            }

            @Override
            public void boosterTemperature(byte[] address, int temperature) {
                Node node = findNode(address);

                if (model.getNode() != null && model.getNode().equals(node)) {
                    LOGGER.trace("Booster temperature has changed, address: {}, temperature: {}", address, temperature);
                    model.setBoosterTemperature(temperature);
                }
            }

            @Override
            public void boosterVoltage(byte[] address, int voltage) {
                Node node = findNode(address);

                if (model.getNode() != null && model.getNode().equals(node)) {
                    LOGGER.trace("Booster voltage has changed, address: {}, voltage: {}", address, voltage);
                    model.setBoosterVoltage(voltage);
                }
            }

            @Override
            public void confidence(byte[] address, int valid, int freeze, int signal) {
                Node node = findNode(address);
                LOGGER.debug("Confidence has changed on node: {}, valid: {}, freeze: {}, signal: {}", node, valid,
                    freeze, signal);

                if (model.getNode() != null && model.getNode().equals(node)) {
                    final List<FeedbackPort> ports = model.getFeedbackPorts();

                    for (int detectorNumber = 0; detectorNumber < ports.size(); detectorNumber++) {
                        model.setFeedbackPortConfidence(detectorNumber, (freeze >> (detectorNumber / 16)) == 1,
                            (signal >> (detectorNumber / 16)) == 1, (valid >> (detectorNumber / 16)) == 1);
                    }
                }
            }

            @Override
            public void free(byte[] address, int detectorNumber) {
                LOGGER.debug("free, address: {}, detectorNumber: {}", address, detectorNumber);
                Node node = findNode(address);

                if (model.getNode() != null && model.getNode().equals(node)) {
                    model.setFeedbackPortStatus(detectorNumber, FeedbackPortStatus.FREE);
                }
                else {
                    LOGGER.debug("Node not found.");
                }
            }

            @Override
            public void identity(byte[] address, IdentifyState identifyState) {
                Node node = findNode(address);

                if (model.getNode() != null && model.getNode().equals(node)) {
                    node.setIdentifyState(identifyState);
                }
            }

            @Override
            public void key(byte[] address, int keyNumber, int keyState) {
                Node node = findNode(address);

                if (model.getNode() != null && model.getNode().equals(node)) {
                    model.setInputPortStatus(keyNumber, keyState);
                }
                else {
                    LOGGER.debug("Node not found.");
                }
            }

            @Override
            public void nodeLost(org.bidib.jbidibc.Node commNode) {
                Node node = findNode(commNode);

                LOGGER.info("Node lost, commNode: {}, node: {}", commNode, node);
                if (node != null) {
                    model.removeNode(node);
                }
                else {
                    LOGGER.debug("Node not found.");
                }
            }

            @Override
            public void nodeNew(final org.bidib.jbidibc.Node commNode) {
                LOGGER.info("New node in system: {}", commNode);
                new Thread(
                    "newNode-created-thread-" + String.valueOf(System.currentTimeMillis())) {
                    @Override
                    public void run() {
                        LOGGER.info("New node created in system: {}", commNode);
                        Communication communication = CommunicationFactory.getInstance();
                        Node node = createNode(communication, commNode);

                        if (node != null) {
                            LOGGER.info("Add new created node to model: {}", node);

                            model.addNode(node);

                            LOGGER.info("sysEnable node: {}", node);
                            communication.enable(node.getNode());
                            LOGGER.info("sysEnable finished, node: {}", node);
                        }
                        LOGGER.info("Create new node has finished: {}", commNode);
                    }
                }.start();
            }

            @Override
            public void occupied(byte[] address, int detectorNumber) {
                LOGGER.debug("occupied, address: {}, detectorNumber: {}", address, detectorNumber);
                Node node = findNode(address);

                LOGGER.debug("occupied, model: {}, node: {}", model, node);
                if (model.getNode() != null && model.getNode().equals(node)) {
                    model.setFeedbackPortStatus(detectorNumber, FeedbackPortStatus.OCCUPIED);
                }
                else {
                    LOGGER.debug("Node not found.");
                }
            }

            @Override
            public void speed(byte[] address, org.bidib.jbidibc.AddressData addressData, int speed) {
                Node node = findNode(address);

                if (model.getNode() != null && model.getNode().equals(node)) {
                    final List<FeedbackPort> ports = model.getFeedbackPorts();

                    for (int detectorNumber = 0; detectorNumber < ports.size(); detectorNumber++) {
                        model.setFeedbackPortSpeed(detectorNumber, addressData.getAddress(), speed);
                    }
                }
                else {
                    LOGGER.debug("Node not found.");
                }
            }
        });
        view = new MainView(model);
        view.setIconImage(new ImageIcon(getClass().getResource("/icons/frameIcon_48x48.png")).getImage());
        view.setVisible(true);

        CommunicationFactory.addCommunicationListener(new CommunicationListener() {

            @Override
            public void opened(String port) {
                view.setStatusText(Resources.getString(CommunicationFactory.class, "open-port-passed") + port, -1);
            }

            @Override
            public void status(String statusText, int displayDuration) {
                view.setStatusText(statusText, displayDuration);
            }
        });

        // get the nodes from the interface
        createNodes(Preferences.getInstance().isMock());

        view.addAccessoryListListener(new AccessoryListListener() {
            @Override
            public void exportAccessory(final Accessory accessory) {
                new FileDialog(
                    FileDialog.SAVE, accessory.toString() + "." + ACCESSORY_EXTENSION, accessoryFilter) {
                    @Override
                    public void approve(String fileName) {
                        try {
                            accessory.save(fileName);
                        }
                        catch (IOException e) {
                            throw new RuntimeException(e);
                        }
                    }
                };
            }

            @Override
            public void importAccessory(Accessory accessory) {
                new FileDialog(
                    FileDialog.OPEN, null, accessoryFilter) {
                    @Override
                    public void approve(String fileName) {
                        Accessory accessory = Accessory.load(fileName);

                        accessory.setId(model.getAccessory().getId());
                        accessory.setMacroSize(CommunicationFactory.getInstance().getAccessoryLength(
                            model.getNode().getNode()));
                        model.replaceAccessory(accessory);
                    }
                };
            }

            @Override
            public void labelChanged(Object object, String label) {
                final Accessory accessory = (Accessory) object;

                accessory.setLabel(label);
                if (label != null && label.length() > 0) {
                    accessoryLabels.setLabel(model.getNode().getNode().getUniqueId(), accessory.getId(), label);
                }
                else {
                    accessoryLabels.removeLabel(model.getNode().getNode().getUniqueId(), accessory.getId());
                }
                try {
                    accessoryLabels.save();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void reloadAccessory(Accessory accessory) {
                CommunicationFactory.getInstance().reloadAccessory(model.getNode().getNode(), accessory);
            }

            @Override
            public void saveAccessory(Accessory accessory) {
                CommunicationFactory.getInstance().saveAccessory(model.getNode().getNode(), accessory);
            }
        });
        view.addAccessoryListSelectionListener(new ListSelectionListener() {
            @Override
            public void valueChanged(final ListSelectionEvent e) {
                if (!e.getValueIsAdjusting()) {
                    setWaitCursor();

                    final JList accessoryList = (JList) e.getSource();
                    Accessory accessory = (Accessory) accessoryList.getSelectedValue();

                    if (accessory != null) {
                        model.setAccessory(accessory);
                    }
                    setDefaultCursor();
                }
            }
        });
        view.addAccessoryTableListener(new AccessoryTableListener() {
            @Override
            public void delete(int[] rows) {
                for (int index = rows.length - 1; index >= 0; index--) {
                    model.getAccessory().removeMacro(rows[index]);
                }
            }

            @Override
            public void insertEmptyAfter(int row) {
                model.getAccessory().addMacroAfter(row, new Macro());
            }

            @Override
            public void insertEmptyBefore(int row) {
                model.getAccessory().addMacroBefore(row >= 0 ? row : 0, new Macro());
            }

            @Override
            public void testButtonPressed(int aspect) {
                CommunicationFactory.getInstance().startAccessory(model.getNode().getNode(), model.getAccessory(),
                    aspect);
            }
        });
        view.addAnalogPortListener(new OutputListener() {
            @Override
            public void labelChanged(Port port, String label) {
                port.setLabel(label);
                if (label != null && label.length() > 0) {
                    analogPortLabels.setLabel(model.getNode().getNode().getUniqueId(), port.getId(), label);
                }
                else {
                    analogPortLabels.removeLabel(model.getNode().getNode().getUniqueId(), port.getId());
                }
                try {
                    analogPortLabels.save();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void statusChanged(Port port, Enum<?> status) {
            }

            @Override
            public void testButtonPressed(Port port) {
                CommunicationFactory.getInstance().activateAnalogPort(model.getNode().getNode(), port.getId(),
                    (AnalogPortStatus) port.getStatus());
            }
        });
        view.addBoosterStatusListener(new StatusListener() {
            @Override
            public void switchedOff() {
                CommunicationFactory.getInstance().boosterOff(model.getNode().getNode());
            }

            @Override
            public void switchedOn() {
                CommunicationFactory.getInstance().boosterOn(model.getNode().getNode());
            }
        });
        view.addInputPortListener(new PortListener() {
            @Override
            public void labelChanged(Port port, String label) {
                port.setLabel(label);
                if (label != null && label.length() > 0) {
                    inputPortLabels.setLabel(model.getNode().getNode().getUniqueId(), port.getId(), label);
                }
                else {
                    inputPortLabels.removeLabel(model.getNode().getNode().getUniqueId(), port.getId());
                }
                try {
                    inputPortLabels.save();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void statusChanged(Port port, Enum<?> status) {
            }
        });
        view.addLightPortListener(new LightPortListener() {
            @Override
            public void valuesChanged(LightPort port) {
                CommunicationFactory.getInstance().setLightPortParameters(model.getNode().getNode(), port.getId(),
                    port.getPwmMin(), port.getPwmMax(), port.getDimMin(), port.getDimMax());
            }

            @Override
            public void labelChanged(Port port, String label) {
                port.setLabel(label);
                if (label != null && label.length() > 0) {
                    lightPortLabels.setLabel(model.getNode().getNode().getUniqueId(), port.getId(), label);
                }
                else {
                    lightPortLabels.removeLabel(model.getNode().getNode().getUniqueId(), port.getId());
                }
                try {
                    lightPortLabels.save();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void statusChanged(Port port, Enum<?> status) {
            }

            @Override
            public void testButtonPressed(Port port) {
                CommunicationFactory.getInstance().activateLightPort(model.getNode().getNode(), port.getId(),
                    (LightPortStatus) port.getStatus());
            }
        });
        view.addMacroListListener(new MacroListListener() {
            @Override
            public void exportMacro(final Macro macro) {
                new FileDialog(
                    FileDialog.SAVE, macro.toString() + "." + MACRO_EXTENSION, macroFilter) {
                    @Override
                    public void approve(String fileName) {
                        try {
                            macro.save(fileName);
                        }
                        catch (IOException e) {
                            throw new RuntimeException(e);
                        }
                    }
                };
            }

            @Override
            public void importMacro() {
                new FileDialog(
                    FileDialog.OPEN, null, macroFilter) {
                    @Override
                    public void approve(String fileName) {
                        Macro macro = Macro.load(fileName);

                        if (macro != null) {
                            replaceReferences(macro);
                            macro.setFunctionSize(CommunicationFactory.getInstance().getMacroLength(
                                model.getNode().getNode()));
                            macro.setId(model.getMacro().getId());
                            model.replaceMacro(macro);
                        }
                    }
                };
            }

            @Override
            public void labelChanged(Object object, String label) {
                final Macro macro = (Macro) object;

                if (label != null && label.length() > 0) {
                    macroLabels.setLabel(model.getNode().getNode().getUniqueId(), macro.getId(), label);
                }
                else {
                    macroLabels.removeLabel(model.getNode().getNode().getUniqueId(), macro.getId());
                }
                try {
                    macroLabels.save();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void reloadMacro(Macro macro) {
                CommunicationFactory.getInstance().reloadMacro(model.getNode().getNode(), macro);
            }

            @Override
            public void saveMacro(Macro macro) {
                CommunicationFactory.getInstance().saveMacro(model.getNode().getNode(), macro);
            }

            @Override
            public void startMacro(Macro macro) {
                CommunicationFactory.getInstance().startMacro(model.getNode().getNode(), macro);
            }

            @Override
            public void stopMacro(Macro macro) {
                CommunicationFactory.getInstance().stopMacro(model.getNode().getNode(), macro);
            }

            @Override
            public void transferMacro(Macro macro) {
                LOGGER.debug("transfer macro: {}", macro);
                CommunicationFactory.getInstance().transferMacro(model.getNode().getNode(), macro);
            }

            @Override
            public void initializeMacro(Macro macro) {
                // initialize the macro
                macro.initialize();
                // and force the panels to reload
                model.setMacro(macro);
            }

        });
        view.addMacroListSelectionListener(new ListSelectionListener() {
            @Override
            public void valueChanged(final ListSelectionEvent e) {
                if (!e.getValueIsAdjusting()) {
                    setWaitCursor();

                    final JList macroList = (JList) e.getSource();
                    Macro macro = (Macro) macroList.getSelectedValue();

                    if (macro != null) {
                        model.setMacro(CommunicationFactory.getInstance().getMacroContent(model.getNode().getNode(),
                            macro));
                    }
                    setDefaultCursor();
                }
            }
        });
        view.addMacroTableListener(new MacroTableListener() {
            @Override
            public void copy(Function[] functions) {
                setFunctions(functions);
            }

            @Override
            public void cut(int[] rows, Function[] functions) {
                setFunctions(functions);
                for (int index = rows.length - 1; index >= 0; index--) {
                    model.getMacro().removeFunction(rows[index]);
                }
            }

            @Override
            public void delete(int[] rows) {
                for (int index = rows.length - 1; index >= 0; index--) {
                    model.getMacro().removeFunction(rows[index]);
                }
            }

            @Override
            public void insertEmptyAfter(int row) {
                model.getMacro().addFunctionsAfter(row, null);
            }

            @Override
            public void insertEmptyBefore(int row) {
                model.getMacro().addFunctionsBefore(row >= 0 ? row : 0, null);
            }

            @Override
            public void pasteAfter(int row) {
                model.getMacro().addFunctionsAfter(row, getFunctions());
            }

            @Override
            public void pasteBefore(int row) {
                model.getMacro().addFunctionsBefore(row >= 0 ? row : 0, getFunctions());
            }
        });
        view.addMainMenuListener(new MainMenuListener() {
            @Override
            public void about() {
                try {
                    new AboutDialog(view);
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void exit() {
                stop();
            }

            @Override
            public void preferences() {
                PreferencesController preferencesController = new PreferencesController(view);

                preferencesController.start();
            }
        });
        view.addMotorPortListener(new OutputListener() {
            @Override
            public void labelChanged(Port port, String value) {
                port.setLabel(value);
                if (value != null && value.length() > 0) {
                    motorPortLabels.setLabel(model.getNode().getNode().getUniqueId(), port.getId(), value);
                }
                else {
                    motorPortLabels.removeLabel(model.getNode().getNode().getUniqueId(), port.getId());
                }
                try {
                    motorPortLabels.save();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void statusChanged(Port port, Enum<?> status) {
            }

            @Override
            public void testButtonPressed(Port port) {
                CommunicationFactory.getInstance().activateMotorPort(model.getNode().getNode(), port.getId(),
                    (MotorPortStatus) port.getStatus());
            }
        });
        view.addNodeListListener(new NodeListListener() {
            @Override
            public void cvProgrammer(Node node, int x, int y) {
                setWaitCursor();

                final CvProgrammerController cvProgrammerController = new CvProgrammerController(view, node, x, y);

                cvProgrammerController.addCvProgrammerControllerListener(new CvProgrammerControllerListener() {
                    @Override
                    public void close() {
                        setDefaultCursor();
                    }

                    @Override
                    public void read(Node node, int cvNumber) {
                        cvProgrammerController.writeCvValue(CommunicationFactory.getInstance().readCv(node.getNode(),
                            cvNumber));
                    }

                    @Override
                    public void write(Node node, int cvNumber, int cvValue) {
                        CommunicationFactory.getInstance().writeCv(node.getNode(), cvNumber, cvValue);
                    }
                });
                cvProgrammerController.start();
            }

            @Override
            public void enableAddressMessages(Node node) {
                CommunicationFactory.getInstance().setAddressMessagesEnabled(node.getNode(),
                    node.isAddressMessagesEnabled());
            }

            @Override
            public void enableDccStart(Node node) {
                CommunicationFactory.getInstance().setDccStartEnabled(node.getNode(), node.isDccStartEnabled());
            }

            @Override
            public void enableExternalStart(Node node) {
                CommunicationFactory.getInstance().setExternalStartEnabled(node.getNode(),
                    node.isExternalStartEnabled());
            }

            @Override
            public void enableFeedbackMessages(Node node) {
                CommunicationFactory.getInstance().setFeedbackMessagesEnabled(node.getNode(),
                    node.isFeedbackMessagesEnabled());
            }

            @Override
            public void enableKeyMessages(Node node) {
                CommunicationFactory.getInstance().setKeyMessagesEnabled(node.getNode(), node.isKeyMessagesEnabled());
            }

            @Override
            public void features(Node node, int x, int y) {
                setWaitCursor();

                final FeaturesController featuresController = new FeaturesController(view, node, x, y);

                featuresController.addFeaturesControllerListener(new FeaturesControllerListener() {
                    @Override
                    public void close() {
                        setDefaultCursor();
                    }

                    @Override
                    public void readAll(Node node) {
                        featuresController.setFeatures(CommunicationFactory.getInstance().getFeatures(node.getNode()));
                    }

                    @Override
                    public void write(Node node, Feature feature) {
                        LOGGER.debug("Write feature to node: {}, feature: {}", node, feature);
                        CommunicationFactory.getInstance().writeFeature(node.getNode(), feature);
                    }
                });
                featuresController.start();
            }

            @Override
            public void exportNode(Node node) {
                new FileDialog(
                    FileDialog.SAVE, node.toString() + "." + NODE_EXTENSION, nodeFilter) {
                    @Override
                    public void approve(String fileName) {
                        try {
                            NodeState nodeState =
                                new NodeState(model.getAccessories(), model.getLightPorts(), model.getMacros(), model
                                    .getServoPorts());

                            nodeState.save(fileName);
                        }
                        catch (IOException e) {
                            throw new RuntimeException(e);
                        }
                    }
                };
            }

            @Override
            public void importNode(Node node) {
                new FileDialog(
                    FileDialog.OPEN, null, nodeFilter) {
                    @Override
                    public void approve(String fileName) {
                        NodeState nodeState = NodeState.load(fileName);

                        for (Accessory accessory : nodeState.getAccessories()) {
                            accessory.setMacroSize(CommunicationFactory.getInstance().getAccessoryLength(
                                model.getNode().getNode()));
                        }
                        model.setAccessories(nodeState.getAccessories());
                        for (Macro macro : nodeState.getMacros()) {
                            replaceReferences(macro);
                            macro.setFunctionSize(CommunicationFactory.getInstance().getMacroLength(
                                model.getNode().getNode()));
                        }
                        model.setMacros(nodeState.getMacros());
                        model.setLightPorts(nodeState.getLightPorts());
                        model.setServoPorts(nodeState.getServoPorts());
                    }
                };
            }

            @Override
            public void firmwareUpdate(Node node, int x, int y) {
                setWaitCursor();

                FirmwareController firmwareController = new FirmwareController(view, node, x, y);

                firmwareController.addFirmwareControllerListener(new FirmwareControllerListener() {
                    @Override
                    public void close() {
                        setDefaultCursor();
                    }
                });
                firmwareController.start();
            }

            @Override
            public void identify(Node node) {
                CommunicationFactory.getInstance().identify(node.getNode(), node.getIdentifyState());
            }

            @Override
            public void labelChanged(Object object, String label) {
                final Node node = (Node) object;

                if (label != null && label.length() > 0) {
                    nodeLabels.setLabel(node.getNode().getUniqueId(), label);
                }
                else {
                    nodeLabels.removeLabel(node.getNode().getUniqueId());
                }
                try {
                    nodeLabels.save();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void loco(Node node, int x, int y) {
                LocoController locoController = new LocoController(node, view, x, y);

                locoController.start();
            }

            @Override
            public void nodeDetails(Node node, int x, int y) {
                // show the node details dialog
                new NodeDetailsDialog(node, CommunicationFactory.getInstance(), x, y);

            }

            @Override
            public void saveNode(Node node) {
                for (Accessory accessory : model.getAccessories()) {
                    CommunicationFactory.getInstance().saveAccessory(node.getNode(), accessory);
                }

                for (Macro macro : model.getMacros()) {
                    CommunicationFactory.getInstance().saveMacro(node.getNode(), macro);
                }

                for (LightPort lightPort : model.getLightPorts()) {
                    CommunicationFactory.getInstance().setLightPortParameters(node.getNode(), lightPort.getId(),
                        lightPort.getPwmMin(), lightPort.getPwmMax(), lightPort.getDimMin(), lightPort.getDimMax());
                }

                for (ServoPort servoPort : model.getServoPorts()) {
                    CommunicationFactory.getInstance().setServoPortParameters(node.getNode(), servoPort.getId(),
                        servoPort.getTrimDown(), servoPort.getTrimUp(), servoPort.getSpeed());
                }
            }
        });
        view.addNodeListSelectionListener(new ListSelectionListener() {
            @Override
            public void valueChanged(final ListSelectionEvent e) {
                if (!e.getValueIsAdjusting()) {
                    LOGGER.debug("The list selection has changed on list: {}", e.getSource());
                    new Thread(
                        "Node-changed-thread-" + String.valueOf(System.currentTimeMillis())) {
                        @Override
                        public void run() {
                            LOGGER.debug("Process node changed starting.");
                            setWaitCursor();

                            final JList nodeList = (JList) e.getSource();
                            final Node node = (Node) nodeList.getSelectedValue();

                            LOGGER.debug("Set the new selected node in the model: {}", node);
                            model.setNode(node);

                            Collection<Accessory> accessories = new LinkedList<Accessory>();
                            Collection<AnalogPort> analogPorts = new LinkedList<AnalogPort>();
                            Collection<Flag> flags = new LinkedList<Flag>();
                            Collection<FeedbackPort> feedbackPorts = new LinkedList<FeedbackPort>();
                            Collection<InputPort> inputPorts = new LinkedList<InputPort>();
                            Collection<LightPort> lightPorts = new LinkedList<LightPort>();
                            Collection<Macro> macros = new LinkedList<Macro>();
                            Collection<MotorPort> motorPorts = new LinkedList<MotorPort>();
                            Collection<ServoPort> servoPorts = new LinkedList<ServoPort>();
                            Collection<SoundPort> soundPorts = new LinkedList<SoundPort>();
                            Collection<SwitchPort> switchPorts = new LinkedList<SwitchPort>();

                            if (node != null && CommunicationFactory.getInstance() != null) {
                                LOGGER.debug("Get the accessories from the node.");
                                Collection<Accessory> nodeAccessories =
                                    CommunicationFactory.getInstance().getAccessories(node.getNode());
                                if (nodeAccessories != null) {
                                    for (Accessory accessory : nodeAccessories) {
                                        accessory.setLabel(accessoryLabels.getLabel(node.getNode().getUniqueId(),
                                            accessory.getId()));
                                        accessories.add(accessory);
                                    }
                                }
                                LOGGER.debug("Get the analog ports from the node.");
                                for (AnalogPort analogPort : CommunicationFactory.getInstance().getAnalogPorts(
                                    node.getNode())) {
                                    analogPort.setLabel(analogPortLabels.getLabel(node.getNode().getUniqueId(),
                                        analogPort.getId()));
                                    analogPorts.add(analogPort);
                                }
                                LOGGER.debug("Get the flags.");
                                for (Flag flag : CommunicationFactory.getInstance().getFlags()) {
                                    flags.add(flag);
                                }
                                LOGGER.debug("Get the feedback ports from the node.");
                                for (FeedbackPort feedbackPort : CommunicationFactory.getInstance().getFeedbackPorts(
                                    node.getNode())) {
                                    feedbackPort.setLabel(feedbackPortLabels.getLabel(node.getNode().getUniqueId(),
                                        feedbackPort.getId()));
                                    feedbackPorts.add(feedbackPort);
                                }
                                LOGGER.debug("Get the input ports from the node.");
                                for (InputPort inputPort : CommunicationFactory.getInstance().getInputPorts(
                                    node.getNode())) {
                                    inputPort.setLabel(inputPortLabels.getLabel(node.getNode().getUniqueId(), inputPort
                                        .getId()));
                                    inputPorts.add(inputPort);
                                }
                                LOGGER.debug("Get the light ports from the node.");
                                for (LightPort lightPort : CommunicationFactory.getInstance().getLightPorts(
                                    node.getNode())) {
                                    lightPort.setLabel(lightPortLabels.getLabel(node.getNode().getUniqueId(), lightPort
                                        .getId()));
                                    lightPorts.add(lightPort);
                                }
                                LOGGER.debug("Get the macros from the node.");
                                for (Macro macro : CommunicationFactory.getInstance().getMacros(node.getNode())) {
                                    macro.setLabel(macroLabels.getLabel(node.getNode().getUniqueId(), macro.getId()));
                                    macros.add(macro);
                                }
                                LOGGER.debug("Get the motor ports from the node.");
                                for (MotorPort motorPort : CommunicationFactory.getInstance().getMotorPorts(
                                    node.getNode())) {
                                    motorPort.setLabel(motorPortLabels.getLabel(node.getNode().getUniqueId(), motorPort
                                        .getId()));
                                    motorPorts.add(motorPort);
                                }
                                LOGGER.debug("Get the servo ports from the node.");
                                for (ServoPort servoPort : CommunicationFactory.getInstance().getServoPorts(
                                    node.getNode())) {
                                    servoPort.setLabel(servoPortLabels.getLabel(node.getNode().getUniqueId(), servoPort
                                        .getId()));
                                    servoPorts.add(servoPort);
                                }
                                LOGGER.debug("Get the sound ports from the node.");
                                for (SoundPort soundPort : CommunicationFactory.getInstance().getSoundPorts(
                                    node.getNode())) {
                                    soundPort.setLabel(soundPortLabels.getLabel(node.getNode().getUniqueId(), soundPort
                                        .getId()));
                                    soundPorts.add(soundPort);
                                }
                                LOGGER.debug("Get the switch ports from the node.");
                                for (SwitchPort switchPort : CommunicationFactory.getInstance().getSwitchPorts(
                                    node.getNode())) {
                                    switchPort.setLabel(switchPortLabels.getLabel(node.getNode().getUniqueId(),
                                        switchPort.getId()));
                                    switchPorts.add(switchPort);
                                }
                            }

                            model.setAccessories(accessories);
                            model.setAnalogPorts(analogPorts);
                            model.setFeedbackPorts(feedbackPorts);
                            model.setFlags(flags);
                            model.setInputPorts(inputPorts);
                            model.setLightPorts(lightPorts);
                            model.setMacros(macros);
                            model.setMotorPorts(motorPorts);
                            model.setServoPorts(servoPorts);
                            model.setSoundPorts(soundPorts);
                            model.setSwitchPorts(switchPorts);

                            // initialize the current values
                            if (node != null && CommunicationFactory.getInstance() != null) {
                                // set the max booster current
                                model.setBoosterMaximumCurrent(CommunicationFactory
                                    .getInstance().getBoosterMaximumCurrent(node.getNode()));

                                if (model.getFeedbackPorts().size() > 0) {
                                    CommunicationFactory.getInstance().getFeedbackPortStatus(node.getNode(), 0,
                                        model.getFeedbackPorts().size());
                                }
                                for (InputPort inputPort : model.getInputPorts()) {
                                    CommunicationFactory.getInstance().getInputPortStatus(node.getNode(),
                                        inputPort.getId());
                                }
                                if (node.getNode().hasBoosterFunctions()) {
                                    CommunicationFactory.getInstance().boosterQuery(node.getNode());
                                }
                            }
                            setDefaultCursor();
                            LOGGER.debug("Process node changed finished.");
                        }
                    }.start();
                }
            }
        });
        view.addServoPortListener(new ServoPortListener() {
            @Override
            public void labelChanged(Port port, String label) {
                port.setLabel(label);
                if (label != null && label.length() > 0) {
                    servoPortLabels.setLabel(model.getNode().getNode().getUniqueId(), port.getId(), label);
                }
                else {
                    servoPortLabels.removeLabel(model.getNode().getNode().getUniqueId(), port.getId());
                }
                try {
                    servoPortLabels.save();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void statusChanged(Port port, Enum<?> status) {
            }

            @Override
            public void testButtonPressed(Port port) {
                CommunicationFactory.getInstance().activateServoPort(model.getNode().getNode(), port.getId(),
                    ((ServoPort) port).getValue());
            }

            @Override
            public void trimDownButtonPressed(Port port) {
                CommunicationFactory.getInstance().activateServoPort(model.getNode().getNode(), port.getId(),
                    ((ServoPort) port).getTrimDown());
            }

            @Override
            public void trimUpButtonPressed(Port port) {
                CommunicationFactory.getInstance().activateServoPort(model.getNode().getNode(), port.getId(),
                    ((ServoPort) port).getTrimUp());
            }

            @Override
            public void valuesChanged(ServoPort port) {
                CommunicationFactory.getInstance().setServoPortParameters(model.getNode().getNode(), port.getId(),
                    port.getTrimDown(), port.getTrimUp(), port.getSpeed());
            }
        });
        view.addSoundPortListener(new OutputListener() {
            @Override
            public void labelChanged(Port port, String label) {
                port.setLabel(label);
                if (label != null && label.length() > 0) {
                    soundPortLabels.setLabel(model.getNode().getNode().getUniqueId(), port.getId(), label);
                }
                else {
                    soundPortLabels.removeLabel(model.getNode().getNode().getUniqueId(), port.getId());
                }
                try {
                    soundPortLabels.save();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void statusChanged(Port port, Enum<?> status) {
            }

            @Override
            public void testButtonPressed(Port port) {
                CommunicationFactory.getInstance().activateSoundPort(model.getNode().getNode(), port.getId(),
                    (SoundPortStatus) port.getStatus());
            }
        });
        view.addStatusListener(new StatusListener() {
            @Override
            public void switchedOff() {
                model.getStatusModel().setRunning(false);
            }

            @Override
            public void switchedOn() {
                model.getStatusModel().setRunning(true);
            }
        });
        view.addSwitchPortListener(new OutputListener() {
            @Override
            public void labelChanged(Port port, String label) {
                port.setLabel(label);
                if (label != null && label.length() > 0) {
                    switchPortLabels.setLabel(model.getNode().getNode().getUniqueId(), port.getId(), label);
                }
                else {
                    switchPortLabels.removeLabel(model.getNode().getNode().getUniqueId(), port.getId());
                }
                try {
                    switchPortLabels.save();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public void statusChanged(Port port, Enum<?> status) {
            }

            @Override
            public void testButtonPressed(Port port) {
                CommunicationFactory.getInstance().activateSwitchPort(model.getNode().getNode(), port.getId(),
                    (SwitchPortStatus) port.getStatus());
            }
        });
        view.setWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                stop();
            }
        });
        Preferences.getInstance().addPreferencesListener(new PreferencesListener() {
            @Override
            public void commPortChanged(String commPort) {
                LOGGER.info("The comm port has changed.");
                createNodes(Preferences.getInstance().isMock());
            }

            @Override
            public void mockChanged(boolean mock) {
                createNodes(mock);
            }

            @Override
            public void startTimeChanged(Date startTime) {
            }

            @Override
            public void timeFactorChanged(int timeFactor) {
            }
        });
    }

    public void stop() {
        view.setVisible(false);
        System.exit(0);
    }
}
