package org.bidib.wizard.mvc.main.model;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;

import org.bidib.wizard.comm.BoosterStatus;
import org.bidib.wizard.comm.FeedbackPortStatus;
import org.bidib.wizard.comm.InputPortStatus;
import org.bidib.wizard.mvc.main.model.listener.AccessoryListListener;
import org.bidib.wizard.mvc.main.model.listener.BoosterStatusListener;
import org.bidib.wizard.mvc.main.model.listener.FeedbackPortListener;
import org.bidib.wizard.mvc.main.model.listener.MacroListListener;
import org.bidib.wizard.mvc.main.model.listener.NodeListListener;
import org.bidib.wizard.mvc.main.model.listener.PortListListener;
import org.bidib.wizard.mvc.main.model.listener.PortListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class MainModel {
    private static final Logger LOGGER = LoggerFactory.getLogger(MainModel.class);

    private Collection<AccessoryListListener> accessoryListListeners = new LinkedList<AccessoryListListener>();

    private Collection<PortListListener> analogPortListListeners = new LinkedList<PortListListener>();

    private Collection<BoosterStatusListener> boosterStatusListeners = new LinkedList<BoosterStatusListener>();

    private Collection<FeedbackPortListener> feedbackPortListeners = new LinkedList<FeedbackPortListener>();

    private Collection<PortListListener> feedbackPortListListeners = new LinkedList<PortListListener>();

    private Collection<PortListener> inputPortListeners = new LinkedList<PortListener>();

    private Collection<PortListListener> inputPortListListeners = new LinkedList<PortListListener>();

    private Collection<PortListListener> lightPortListListeners = new LinkedList<PortListListener>();

    private Collection<MacroListListener> macroListListeners = new LinkedList<MacroListListener>();

    private Collection<PortListListener> motorPortListListeners = new LinkedList<PortListListener>();

    private Collection<NodeListListener> nodeListListeners = new LinkedList<NodeListListener>();

    private Collection<PortListListener> servoPortListListeners = new LinkedList<PortListListener>();

    private Collection<PortListListener> soundPortListListeners = new LinkedList<PortListListener>();

    private Collection<PortListListener> switchPortListListeners = new LinkedList<PortListListener>();

    private List<Accessory> accessories = new LinkedList<Accessory>();

    private List<AnalogPort> analogPorts = new LinkedList<AnalogPort>();

    private int boosterCurrent = 0;

    private int boosterMaximumCurrent = 0;

    private BoosterStatus boosterStatus = null;

    private int boosterTemperature = 0;

    private int boosterVoltage = 0;

    private List<FeedbackPort> feedbackPorts = new LinkedList<FeedbackPort>();

    private List<Flag> flags = new LinkedList<Flag>();

    private List<InputPort> inputPorts = new LinkedList<InputPort>();

    private List<LightPort> lightPorts = new LinkedList<LightPort>();

    private List<Macro> macros = new LinkedList<Macro>();

    private List<MotorPort> motorPorts = new LinkedList<MotorPort>();

    private Collection<Node> nodes = new LinkedHashSet<Node>();

    private List<ServoPort> servoPorts = new LinkedList<ServoPort>();

    private List<SoundPort> soundPorts = new LinkedList<SoundPort>();

    private List<SwitchPort> switchPorts = new LinkedList<SwitchPort>();

    private Accessory accessory = null;

    private Macro macro = null;

    private Node node = null;

    private StatusModel statusModel = new StatusModel();

    public void addAccessoryListListener(AccessoryListListener l) {
        accessoryListListeners.add(l);
    }

    public void addAnalogPortListListener(PortListListener l) {
        analogPortListListeners.add(l);
    }

    public void addBoosterStatusListener(BoosterStatusListener l) {
        boosterStatusListeners.add(l);
    }

    public void addFeedbackPortListener(FeedbackPortListener l) {
        feedbackPortListeners.add(l);
    }

    public void addFeedbackPortListListener(PortListListener l) {
        feedbackPortListListeners.add(l);
    }

    public void addInputPortListener(PortListener l) {
        inputPortListeners.add(l);
    }

    public void addInputPortListListener(PortListListener l) {
        inputPortListListeners.add(l);
    }

    public void addLightPortListListener(PortListListener l) {
        lightPortListListeners.add(l);
    }

    public void addMacroListListener(MacroListListener l) {
        macroListListeners.add(l);
    }

    public void addMotorPortListListener(PortListListener l) {
        motorPortListListeners.add(l);
    }

    /**
     * Add a node to the list of nodes.
     * @param node the new node to add
     */
    public void addNode(Node node) {
        boolean nodeListChanged = false;
        synchronized (nodes) {
            LOGGER.debug("Add new node: {}", node);
            if (nodes.add(node)) {
                //                fireNodeListChanged();
                nodeListChanged = true;
            }
        }
        if (nodeListChanged) {
            fireNodeListChanged();
        }
    }

    public void addNodeListListener(NodeListListener l) {
        nodeListListeners.add(l);
    }

    public void addServoPortListListener(PortListListener l) {
        servoPortListListeners.add(l);
    }

    public void addSoundPortListListener(PortListListener l) {
        soundPortListListeners.add(l);
    }

    public void addSwitchPortListListener(PortListListener l) {
        switchPortListListeners.add(l);
    }

    private void fireAccessoryChanged() {
        for (AccessoryListListener l : accessoryListListeners) {
            l.accessoryChanged();
        }
    }

    private void fireAccessoryListChanged() {
        for (AccessoryListListener l : accessoryListListeners) {
            l.listChanged();
        }
    }

    private void fireAnalogPortListChanged() {
        for (PortListListener l : analogPortListListeners) {
            l.listChanged();
        }
    }

    private void fireBoosterCurrentChanged(int current) {
        for (BoosterStatusListener l : boosterStatusListeners) {
            l.currentChanged(current);
        }
    }

    private void fireBoosterMaximumCurrentChanged(int maximumCurrent) {
        for (BoosterStatusListener l : boosterStatusListeners) {
            l.maximumCurrentChanged(maximumCurrent);
        }
    }

    private void fireBoosterStatusChanged(BoosterStatus status) {
        for (BoosterStatusListener l : boosterStatusListeners) {
            l.stateChanged(status);
        }
    }

    private void fireBoosterTemperatureChanged(int temperature) {
        for (BoosterStatusListener l : boosterStatusListeners) {
            l.temperatureChanged(temperature);
        }
    }

    private void fireBoosterVoltageChanged(int voltage) {
        for (BoosterStatusListener l : boosterStatusListeners) {
            l.voltageChanged(voltage);
        }
    }

    private void fireFeedbackPortAddressesChanged(FeedbackPort port, Collection<AddressData> addresses) {
        for (FeedbackPortListener l : feedbackPortListeners) {
            l.addressesChanged(port, addresses);
        }
    }

    private void fireFeedbackPortConfidenceChanged(FeedbackPort port, boolean freeze, boolean signal, boolean valid) {
        for (FeedbackPortListener l : feedbackPortListeners) {
            l.confidenceChanged(port, freeze, signal, valid);
        }
    }

    private void fireFeedbackPortListChanged() {
        for (PortListListener l : feedbackPortListListeners) {
            l.listChanged();
        }
    }

    private void fireFeedbackPortSpeedChanged(FeedbackPort port, int address, int speed) {
        for (FeedbackPortListener l : feedbackPortListeners) {
            l.speedChanged(port, address, speed);
        }
    }

    private void fireFeedbackPortStatusChanged(final FeedbackPort port, final FeedbackPortStatus status) {
        for (PortListener l : feedbackPortListeners) {
            l.statusChanged(port, status);
        }
    }

    private void fireFlagListChanged() {
    }

    private void fireInputPortStatusChanged(InputPort port, InputPortStatus status) {
        for (PortListener l : inputPortListeners) {
            l.statusChanged(port, status);
        }
    }

    private void fireInputPortListChanged() {
        for (PortListListener l : inputPortListListeners) {
            l.listChanged();
        }
    }

    private void fireLightPortListChanged() {
        for (PortListListener l : lightPortListListeners) {
            l.listChanged();
        }
    }

    private void fireMacroChanged() {
        for (MacroListListener l : macroListListeners) {
            l.macroChanged();
        }
    }

    private void fireMacroListChanged() {
        for (MacroListListener l : macroListListeners) {
            l.listChanged();
        }
    }

    private void fireMotorPortListChanged() {
        for (PortListListener l : motorPortListListeners) {
            l.listChanged();
        }
    }

    private void fireNodeChanged() {
        for (NodeListListener l : nodeListListeners) {
            l.nodeChanged();
        }
    }

    private void fireNodeListChanged() {
        for (NodeListListener l : nodeListListeners) {
            l.listChanged();
        }
    }

    private void fireServoPortListChanged() {
        for (PortListListener l : servoPortListListeners) {
            l.listChanged();
        }
    }

    private void fireSoundPortListChanged() {
        for (PortListListener l : soundPortListListeners) {
            l.listChanged();
        }
    }

    private void fireSwitchPortListChanged() {
        for (PortListListener l : switchPortListListeners) {
            l.listChanged();
        }
    }

    public Accessory getAccessory() {
        return accessory;
    }

    public List<Accessory> getAccessories() {
        return Collections.unmodifiableList(accessories);
    }

    public List<AnalogPort> getAnalogPorts() {
        return Collections.unmodifiableList(analogPorts);
    }

    public int getBoosterCurrent() {
        return boosterCurrent;
    }

    public int getBoosterMaximumCurrent() {
        return boosterMaximumCurrent;
    }

    public BoosterStatus getBoosterStatus() {
        return boosterStatus;
    }

    public int getBoosterTemperature() {
        return boosterTemperature;
    }

    public int getBoosterVoltage() {
        return boosterVoltage;
    }

    public List<FeedbackPort> getFeedbackPorts() {
        synchronized (feedbackPorts) {
            return Collections.unmodifiableList(feedbackPorts);
        }
    }

    public List<Flag> getFlags() {
        return Collections.unmodifiableList(flags);
    }

    public List<InputPort> getInputPorts() {
        return Collections.unmodifiableList(inputPorts);
    }

    public List<LightPort> getLightPorts() {
        return Collections.unmodifiableList(lightPorts);
    }

    public Macro getMacro() {
        return macro;
    }

    public List<Macro> getMacros() {
        return Collections.unmodifiableList(macros);
    }

    public List<MotorPort> getMotorPorts() {
        return Collections.unmodifiableList(motorPorts);
    }

    public Node getNode() {
        return node;
    }

    public Collection<Node> getNodes() {
        LOGGER.trace("Get the nodes.");
        return Collections.unmodifiableCollection(nodes);
    }

    public List<ServoPort> getServoPorts() {
        return Collections.unmodifiableList(servoPorts);
    }

    public List<SoundPort> getSoundPorts() {
        return Collections.unmodifiableList(soundPorts);
    }

    public StatusModel getStatusModel() {
        return statusModel;
    }

    public List<SwitchPort> getSwitchPorts() {
        return Collections.unmodifiableList(switchPorts);
    }

    public void removeNode(Node node) {
        boolean nodeListChanged = false;
        synchronized (nodes) {
            LOGGER.debug("Remove node: {}", node);
            if (nodes.remove(node)) {
                LOGGER.debug("Removed node: {}", node);
                //                fireNodeListChanged();
                nodeListChanged = true;
            }
            else {
                LOGGER.warn("Remove node from nodes failed: {}", node);
            }
        }
        if (nodeListChanged) {
            fireNodeListChanged();
        }
    }

    public void replaceAccessory(Accessory accessory) {
        accessories.set(accessory.getId(), accessory);
        fireAccessoryListChanged();
        setAccessory(accessory);
    }

    public void replaceMacro(Macro macro) {
        macros.set(macro.getId(), macro);
        fireMacroListChanged();
        setMacro(macro);
    }

    public void setAccessory(Accessory accessory) {
        this.accessory = accessory;
        fireAccessoryChanged();
    }

    public void setAccessories(Collection<Accessory> accessories) {
        this.accessories.clear();
        this.accessories.addAll(accessories);
        fireAccessoryListChanged();
    }

    public void setAnalogPorts(Collection<AnalogPort> analogPorts) {
        this.analogPorts.clear();
        this.analogPorts.addAll(analogPorts);
        fireAnalogPortListChanged();
    }

    public void setBoosterCurrent(int current) {
        if (current != this.boosterCurrent) {
            LOGGER.debug("Booster current has changed, new: {}, old: {}", current, boosterCurrent);
            this.boosterCurrent = current;
            fireBoosterCurrentChanged(current);
        }
        else {
            LOGGER.trace("Current is equal, do not send change notification.");
        }
    }

    public void setBoosterMaximumCurrent(int maximumCurrent) {
        if (maximumCurrent != this.boosterMaximumCurrent) {
            this.boosterMaximumCurrent = maximumCurrent;
            fireBoosterMaximumCurrentChanged(maximumCurrent);
        }
    }

    public void setBoosterStatus(BoosterStatus status) {
        if (status != this.boosterStatus) {
            this.boosterStatus = status;
            fireBoosterStatusChanged(status);
        }
    }

    public void setBoosterTemperature(int temperature) {
        if (temperature != this.boosterTemperature) {
            this.boosterTemperature = temperature;
            fireBoosterTemperatureChanged(temperature);
        }
    }

    public void setBoosterVoltage(int voltage) {
        if (voltage != this.boosterVoltage) {
            this.boosterVoltage = voltage;
            fireBoosterVoltageChanged(voltage);
        }
    }

    public void setFeedbackPortAddresses(int detectorNumber, Collection<AddressData> addresses) {
        if (detectorNumber < getFeedbackPorts().size()) {
            FeedbackPort port = getFeedbackPorts().get(detectorNumber);

            port.setAddresses(addresses);
            fireFeedbackPortAddressesChanged(port, addresses);
        }
    }

    public void setFeedbackPortConfidence(int detectorNumber, boolean freeze, boolean signal, boolean valid) {
        final List<FeedbackPort> ports = getFeedbackPorts();

        for (int index = 0; index < ports.size(); index++) {
            final FeedbackPort port = ports.get(index);

            port.setFreeze(freeze);
            port.setSignal(signal);
            port.setValid(valid);
            fireFeedbackPortConfidenceChanged(port, freeze, signal, valid);
        }
    }

    public void setFeedbackPorts(Collection<FeedbackPort> feedbackPorts) {
        LOGGER.debug("Set the feedback ports, size: {}", (feedbackPorts != null ? feedbackPorts.size() : 0));
        synchronized (this.feedbackPorts) {
            this.feedbackPorts.clear();
            this.feedbackPorts.addAll(feedbackPorts);
        }
        fireFeedbackPortListChanged();
    }

    public void setFeedbackPortSpeed(int detectorNumber, int address, int speed) {
        if (detectorNumber < getFeedbackPorts().size()) {
            FeedbackPort port = getFeedbackPorts().get(detectorNumber);

            if (port != null) {
                Collection<AddressData> addresses = port.getAddresses();

                if (addresses != null) {
                    for (AddressData addressData : addresses) {
                        if (address == addressData.getAddress()) {
                            addressData.setSpeed(speed);
                            fireFeedbackPortSpeedChanged(port, address, speed);
                        }
                    }
                }
            }
        }
    }

    public void setFeedbackPortStatus(int detectorNumber, FeedbackPortStatus status) {
        LOGGER.debug("feedback port status, detector number: {}, status: {}", detectorNumber, status);
        if (detectorNumber < getFeedbackPorts().size()) {
            FeedbackPort port = getFeedbackPorts().get(detectorNumber);

            port.setStatus(status);
            fireFeedbackPortStatusChanged(port, status);
        }
    }

    public void setFlags(Collection<Flag> flags) {
        this.flags.clear();
        this.flags.addAll(flags);
        fireFlagListChanged();
    }

    public void setInputPorts(Collection<InputPort> inputPorts) {
        this.inputPorts.clear();
        this.inputPorts.addAll(inputPorts);
        fireInputPortListChanged();
    }

    public void setInputPortStatus(int keyNumber, int keyState) {
        if (getInputPorts().size() > keyNumber) {
            InputPort port = getInputPorts().get(keyNumber);
            InputPortStatus status = keyState == 1 ? InputPortStatus.ON : InputPortStatus.OFF;

            port.setStatus(status);
            fireInputPortStatusChanged(port, status);
        }
    }

    public void setLightPorts(Collection<LightPort> lightPorts) {
        this.lightPorts.clear();
        this.lightPorts.addAll(lightPorts);
        fireLightPortListChanged();
    }

    public void setMacro(Macro macro) {
        this.macro = macro;
        fireMacroChanged();
    }

    public void setMacros(Collection<Macro> macros) {
        this.macros.clear();
        this.macros.addAll(macros);
        fireMacroListChanged();
    }

    public void setMotorPorts(Collection<MotorPort> motorPorts) {
        this.motorPorts.clear();
        this.motorPorts.addAll(motorPorts);
        fireMotorPortListChanged();
    }

    public void setNode(Node node) {
        LOGGER.debug("Set the models node: {}", node);
        this.node = node;
        fireNodeChanged();
    }

    public void setNodes(Collection<Node> nodes) {
        LOGGER.debug("Set the nodes: {}", nodes);
        synchronized (nodes) {
            this.nodes.clear();
            this.nodes.addAll(nodes);
        }
        fireNodeListChanged();
    }

    public void setServoPorts(Collection<ServoPort> servoPorts) {
        this.servoPorts.clear();
        this.servoPorts.addAll(servoPorts);
        fireServoPortListChanged();
    }

    public void setSoundPorts(Collection<SoundPort> soundPorts) {
        this.soundPorts.clear();
        this.soundPorts.addAll(soundPorts);
        fireSoundPortListChanged();
    }

    public void setSwitchPorts(Collection<SwitchPort> switchPorts) {
        this.switchPorts.clear();
        this.switchPorts.addAll(switchPorts);
        fireSwitchPortListChanged();
    }
}
