package org.bidib.wizard.mvc.main.view.panel;

import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Calendar;
import java.util.Date;

import javax.swing.BorderFactory;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JSpinner;
import javax.swing.SpinnerDateModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.bidib.wizard.locale.Resources;
import org.bidib.wizard.mvc.common.view.panel.DisabledPanel;
import org.bidib.wizard.mvc.main.model.Macro;
import org.bidib.wizard.mvc.main.model.MacroRepeatDay;
import org.bidib.wizard.mvc.main.model.MacroRepeatTime;
import org.bidib.wizard.mvc.main.model.MainModel;
import org.bidib.wizard.mvc.main.model.StartCondition;
import org.bidib.wizard.mvc.main.model.TimeStartCondition;
import org.bidib.wizard.mvc.main.model.listener.MacroListListener;

public class MacroParameterPanel extends JPanel {
    private static final long serialVersionUID = -779625516620733046L;

    private final JSpinner timeSpinner = new JSpinner(new SpinnerDateModel());

    private final DisabledPanel disabledRepeatPanel;

    public MacroParameterPanel(final MainModel model) {
        setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Resources.getString(getClass(),
            "title")
            + ":"));
        setLayout(new GridBagLayout());

        final JPanel repeatPanel = new JPanel(new GridLayout(2, 1));

        final JComboBox repeatTime = new JComboBox();
        repeatTime.setModel(new DefaultComboBoxModel(MacroRepeatTime.values()));
        final JComboBox repeatDay = new JComboBox();
        repeatDay.setModel(new DefaultComboBoxModel(MacroRepeatDay.values()));

        timeSpinner.addChangeListener(new ChangeListener() {
            @Override
            public void stateChanged(ChangeEvent e) {
                for (StartCondition startCondition : model.getMacro().getStartConditions()) {
                    if (startCondition instanceof TimeStartCondition) {
                        JSpinner spinner = (JSpinner) e.getSource();
                        Calendar spinnerTime = Calendar.getInstance();

                        spinnerTime.setTime((Date) spinner.getValue());

                        Calendar time = Calendar.getInstance();

                        time.set(Calendar.HOUR_OF_DAY, spinnerTime.get(Calendar.HOUR_OF_DAY));
                        time.set(Calendar.MINUTE, spinnerTime.get(Calendar.MINUTE));
                        ((TimeStartCondition) startCondition).setTime(time);
                    }
                }
            }
        });

        final JCheckBox timeButton = new JCheckBox(Resources.getString(getClass(), "time") + ":");

        timeButton.addItemListener(new ItemListener() {
            @Override
            public void itemStateChanged(ItemEvent e) {
                timeSpinner.setEnabled(e.getStateChange() == ItemEvent.SELECTED);
                disabledRepeatPanel.setEnabled(e.getStateChange() == ItemEvent.SELECTED);
                if (e.getStateChange() == ItemEvent.SELECTED) {
                    Calendar time = Calendar.getInstance();

                    time.setTime((Date) timeSpinner.getValue());
                    model.getMacro().addStartCondition(new TimeStartCondition(time));
                }
                else {
                    model.getMacro().removeStartCondition(new TimeStartCondition());
                }
            }
        });

        JPanel startConditionPanel = new JPanel(new GridBagLayout());

        startConditionPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Resources
            .getString(getClass(), "condition")
            + ":"));

        FlowLayout leftLayout = new FlowLayout();

        leftLayout.setAlignment(FlowLayout.LEFT);
        timeSpinner.setEditor(new JSpinner.DateEditor(timeSpinner, "HH:mm"));

        JPanel timePanel = new JPanel(leftLayout);

        timePanel.add(timeButton);
        timePanel.add(timeSpinner);

        GridBagConstraints c = new GridBagConstraints();

        c.anchor = GridBagConstraints.FIRST_LINE_START;
        c.fill = GridBagConstraints.NONE;
        c.gridx = 0;
        c.gridy = 0;
        c.insets = new Insets(5, 5, 5, 5);
        c.weightx = 0;
        c.weighty = 0;

        startConditionPanel.add(timePanel, c);

        repeatTime.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                for (StartCondition startCondition : model.getMacro().getStartConditions()) {
                    if (startCondition instanceof TimeStartCondition) {
                        MacroRepeatTime repeatTime = (MacroRepeatTime) ((JComboBox) e.getSource()).getSelectedItem();

                        ((TimeStartCondition) startCondition).setRepeatTime(repeatTime);
                        repeatDay.setEnabled(repeatTime != MacroRepeatTime.NONE);
                    }
                }
            }
        });
        repeatPanel.add(repeatTime);
        repeatDay.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                for (StartCondition startCondition : model.getMacro().getStartConditions()) {
                    if (startCondition instanceof TimeStartCondition) {
                        ((TimeStartCondition) startCondition).setRepeatDay((MacroRepeatDay) ((JComboBox) e.getSource())
                            .getSelectedItem());
                    }
                }
            }
        });
        repeatPanel.add(repeatDay);
        repeatPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Resources.getString(
            getClass(), "repetition")
            + ":"));
        c.fill = GridBagConstraints.BOTH;
        c.gridy++;
        c.weighty = 1;
        startConditionPanel.add(repeatPanel, c);
        disabledRepeatPanel = new DisabledPanel(repeatPanel);
        startConditionPanel.add(disabledRepeatPanel, c);

        c.gridy = 0;
        add(startConditionPanel, c);

        final SliderPanel delayPanel = new SliderPanel(
            Resources.getString(getClass(), "delay") + ":", 1, 255, 255, true) {
            private static final long serialVersionUID = 4415724813270973012L;

            @Override
            public void stateChanged(ChangeEvent e) {
                Macro macro = model.getMacro();

                if (macro != null) {
                    JSlider source = (JSlider) e.getSource();

                    macro.setSpeed((int) source.getValue());
                    setLabel(macro.getSpeed());
                }
            }
        };

        c.anchor = GridBagConstraints.FIRST_LINE_END;
        c.gridx++;
        c.weightx = 1;
        add(delayPanel, c);

        final SliderPanel cyclesPanel = new SliderPanel(
            Resources.getString(getClass(), "cycles") + ":", 0, 250, 1, false) {
            private static final long serialVersionUID = 2568614980702327505L;

            @Override
            public void stateChanged(ChangeEvent e) {
                Macro macro = model.getMacro();

                if (macro != null) {
                    JSlider source = (JSlider) e.getSource();

                    macro.setCycles((int) source.getValue());
                    setLabel(macro.getCycles());
                }
            }
        };

        c.gridx++;
        add(cyclesPanel, c);

        model.addMacroListListener(new MacroListListener() {
            @Override
            public void listChanged() {
            }

            @Override
            public void macroChanged() {
                Macro macro = model.getMacro();

                if (macro != null) {
                    delayPanel.setValue(macro.getSpeed());
                    cyclesPanel.setValue(macro.getCycles());

                    boolean hasTimeCondition = false;

                    for (StartCondition startCondition : model.getMacro().getStartConditions()) {
                        if (startCondition instanceof TimeStartCondition) {
                            hasTimeCondition = true;

                            Calendar time = ((TimeStartCondition) startCondition).getTime();

                            if (time != null) {
                                timeSpinner.setValue(time.getTime());
                            }
                            repeatDay
                                .setEnabled(((TimeStartCondition) startCondition).getRepeatTime() != MacroRepeatTime.NONE);
                            repeatDay.setSelectedIndex(((TimeStartCondition) startCondition).getRepeatDay().ordinal());
                            repeatTime
                                .setSelectedIndex(((TimeStartCondition) startCondition).getRepeatTime().ordinal());
                        }
                    }
                    timeButton.setSelected(hasTimeCondition);
                    timeSpinner.setEnabled(hasTimeCondition);
                    disabledRepeatPanel.setEnabled(hasTimeCondition);
                }
            }
        });
    }
}
